# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
#
from __future__ import annotations

import pytest

import awkward as ak


def test():
    ak.concatenate([ak.to_regular([[i, i], [i, i]]) for i in range(64)], axis=1)
    ak.concatenate([ak.to_regular([[i, i], [i, i]]) for i in range(127)], axis=1)
    ak.concatenate([ak.to_regular([[i, i], [i, i]]) for i in range(128)], axis=1)
    ak.concatenate([ak.to_regular([[i, i], [i, i]]) for i in range(129)], axis=1)
    ak.concatenate([ak.to_regular([[i, i], [i, i]]) for i in range(256)], axis=1)

    ak.concatenate(
        [ak.to_regular([[i, str(i)], [str(i), i]]) for i in range(64)], axis=1
    )
    ak.concatenate(
        [ak.to_regular([[i, str(i)], [str(i), i]]) for i in range(127)], axis=1
    )
    ak.concatenate(
        [ak.to_regular([[i, str(i)], [str(i), i]]) for i in range(128)], axis=1
    )
    ak.concatenate(
        [ak.to_regular([[i, str(i)], [str(i), i]]) for i in range(129)], axis=1
    )

    ak.concatenate(
        [ak.to_regular([[i, str(i)], [str(i), i]]) for i in range(64)], axis=0
    )
    with pytest.raises(
        ValueError, match="UnionArray cannot have more than 128 content types"
    ):
        ak.concatenate(
            [ak.to_regular([[i, str(i)], [str(i), i]]) for i in range(65)], axis=0
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from collections.abc import Callable, Iterator

import awkward as ak
from awkward._meta.unmaskedmeta import UnmaskedMeta
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._parameters import (
    parameters_union,
)
from awkward._typing import Any, DType, Self, final
from awkward._util import UNSET
from awkward.forms.form import Form, _SpecifierMatcher

__all__ = ("UnmaskedForm",)

np = NumpyMetadata.instance()


@final
class UnmaskedForm(UnmaskedMeta[Form], Form):
    _content: Form

    def __init__(
        self,
        content,
        *,
        parameters=None,
        form_key=None,
    ):
        if not isinstance(content, Form):
            raise TypeError(
                f"{type(self).__name__} all 'contents' must be Form subclasses, not {content!r}"
            )

        self._content = content
        self._init(parameters=parameters, form_key=form_key)

    @property
    def content(self):
        return self._content

    def copy(
        self,
        content=UNSET,
        *,
        parameters=UNSET,
        form_key=UNSET,
    ):
        return UnmaskedForm(
            self._content if content is UNSET else content,
            parameters=self._parameters if parameters is UNSET else parameters,
            form_key=self._form_key if form_key is UNSET else form_key,
        )

    @classmethod
    def simplified(
        cls,
        content,
        *,
        parameters=None,
        form_key=None,
    ):
        if content.is_union:
            return content.copy(
                contents=[cls.simplified(x) for x in content.contents],
                parameters=parameters_union(content._parameters, parameters),
            )
        elif content.is_indexed or content.is_option:
            return content.copy(
                parameters=parameters_union(content._parameters, parameters)
            )
        else:
            return cls(content, parameters=parameters, form_key=form_key)

    def __repr__(self):
        args = [repr(self._content), *self._repr_args()]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _to_dict_part(self, verbose, toplevel):
        return self._to_dict_extra(
            {
                "class": "UnmaskedArray",
                "content": self._content._to_dict_part(verbose, toplevel=False),
            },
            verbose,
        )

    @property
    def type(self):
        return ak.types.OptionType(
            self._content.type,
            parameters=self._parameters,
        ).simplify_option_union()

    def _columns(self, path, output, list_indicator):
        self._content._columns(path, output, list_indicator)

    def _prune_columns(self, is_inside_record_or_union: bool) -> Self | None:
        next_content = self._content._prune_columns(is_inside_record_or_union)
        if next_content is None:
            return None
        else:
            return self.copy(content=next_content)

    def _select_columns(self, match_specifier: _SpecifierMatcher) -> Self:
        return self.copy(content=self._content._select_columns(match_specifier))

    def _column_types(self):
        return self._content._column_types()

    def __setstate__(self, state):
        if isinstance(state, dict):
            # read data pickled in Awkward 2.x
            self.__dict__.update(state)
        else:
            # read data pickled in Awkward 1.x

            # https://github.com/scikit-hep/awkward/blob/main-v1/src/python/forms.cpp#L792-L797
            _has_identities, parameters, form_key, content = state

            if form_key is not None:
                form_key = "part0-" + form_key  # only the first partition

            self.__init__(content, parameters=parameters, form_key=form_key)

    def _expected_from_buffers(
        self, getkey: Callable[[Form, str], str], recursive: bool
    ) -> Iterator[tuple[str, DType]]:
        if recursive:
            yield from self._content._expected_from_buffers(getkey, recursive)

    def _is_equal_to(self, other: Any, all_parameters: bool, form_key: bool) -> bool:
        return self._is_equal_to_generic(
            other, all_parameters, form_key
        ) and self._content._is_equal_to(other._content, all_parameters, form_key)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

# taken from test_2719_typetracer_buffer_key.py
form = ak.forms.from_dict(
    {
        "class": "RecordArray",
        "fields": ["x"],
        "contents": [
            {
                "class": "ListOffsetArray",
                "offsets": "i64",
                "content": {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "inner_shape": [],
                    "parameters": {},
                    "form_key": "x.list.content",
                },
                "parameters": {},
                "form_key": "x.list",
            }
        ],
        "parameters": {},
    }
)


def test_buffer_keys_on_virtual_arrays():
    buffers = {
        "x.list-offsets": lambda: np.array([0, 2, 3, 5]),
        "x.list.content-data": lambda: np.array([1, 2, 3, 4, 5]),
    }

    va = ak.from_buffers(form, 3, buffers, buffer_key="{form_key}-{attribute}")

    assert va.layout.content("x").offsets.data.buffer_key == "x.list-offsets"
    assert va.layout.content("x").content.data.buffer_key == "x.list.content-data"


def test_buffer_keys_on_placeholder_arrays():
    buffers = {
        "x.list-offsets": ak._nplikes.placeholder.PlaceholderArray(
            shape=(4,),
            dtype=np.int64,
            buffer_key="x.list-offsets",
            nplike=ak._nplikes.numpy.Numpy.instance(),
        ),
        "x.list.content-data": ak._nplikes.placeholder.PlaceholderArray(
            shape=(5,),
            dtype=np.int64,
            buffer_key="x.list.content-data",
            nplike=ak._nplikes.numpy.Numpy.instance(),
        ),
    }

    pa = ak.from_buffers(form, 3, buffers, buffer_key="{form_key}-{attribute}")

    assert pa.layout.content("x").offsets.data.buffer_key == "x.list-offsets"
    assert pa.layout.content("x").content.data.buffer_key == "x.list.content-data"

    with pytest.raises(
        RuntimeError,
        match=r"Awkward Array tried to access a buffer at 'x.list-offsets'",
    ):
        pa.layout.content("x").offsets.data.materialize()

    with pytest.raises(
        RuntimeError,
        match=r"Awkward Array tried to access a buffer at 'x.list.content-data'",
    ):
        pa.layout.content("x").content.data.materialize()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak
from awkward.operations import to_list


def test_ak_where_with_optional_unknowns():
    """
    This is the example from the Issue.
    In the two cases we fail, the not-selected value has type ?unknown, value None of course.

    Names here are changed a little from github issue 3098.
    """
    opt_true_cond = ak.Array([[True], [None]])[0]  # <Array [True] type='1 * ?bool'>
    true_cond = ak.Array([True])  # <Array [True] type='1 * bool'>
    none_alternative = ak.Array([None])  # <Array [None] type='1 * ?unknown'>
    zero_alternative = ak.Array([0])  # <Array [0] type='1 * int64'>
    opt_zero_alternative = ak.Array([[0], [None]])[0]  # <Array [0] type='1 * ?int64'>

    assert ak.where(opt_true_cond, 1, none_alternative).to_list() == [1]
    # Above fails at time of writing. We get [None].
    assert ak.where(opt_true_cond, 1, zero_alternative).to_list() == [1]
    assert ak.where(opt_true_cond, 1, opt_zero_alternative).to_list() == [1]

    # These assertions pass. Note that true_cond is type bool, not ?bool.
    assert ak.where(true_cond, 1, none_alternative).to_list() == [1]
    assert ak.where(true_cond, 1, zero_alternative).to_list() == [1]
    assert ak.where(true_cond, 1, opt_zero_alternative).to_list() == [1]

    # Like the first three assertions, The first one here fails.
    # This demonstrates that the problem is symmetric w/rt X and Y arrays.
    assert ak.where(~opt_true_cond, none_alternative, 1).to_list() == [1]
    # Above fails at time of writing. We're getting [None] again.
    assert ak.where(~opt_true_cond, zero_alternative, 1).to_list() == [1]
    assert ak.where(~opt_true_cond, opt_zero_alternative, 1).to_list() == [1]


def test_ak_where_with_optionals():
    """
    It turns out that we don't need to use ?unknown arrays to trigger this issue.
    We only need a None (masked value) in an element that is selected against.

    At time-of-writing (ATOW), ak.where() produces None values when:
    1. The conditional array values have OptionType (at least ?bool or ?int64), *AND*
    2. The value array element *NOT* selected has Option type and holds a None value.
    In this case regardless of the type or value of the array element that *IS* selected,
    the result for that element will, incorrectly, be None.
    """
    # This passes. Note that a condition of None creates a None in the result.
    assert to_list(
        ak.where(
            ak.Array([True, False, None]), ak.Array([1, 2, 3]), ak.Array([4, 5, 6])
        )
    ) == [1, 5, None]

    # This also passes. (The presence of None at the end forces option types to be used.)
    assert to_list(
        ak.where(
            ak.Array([True, False, None]),
            ak.Array([1, 2, None]),
            ak.Array([4, 5, None]),
        )
    ) == [1, 5, None]

    # This fails (ATOW). The presence of None forces option types to be used.
    assert to_list(
        ak.where(
            ak.Array([True, False, None]),
            ak.Array([1, 2, None]),
            ak.Array([None, 5, None]),
        )
    ) == [1, 5, None]  # ATOW we get [None, 5, None]

    # Fails ATOW. Same as above but with a None in the X argument.
    assert to_list(
        ak.where(
            ak.Array([True, False, None]),
            ak.Array([1, None, None]),
            ak.Array([4, 5, None]),
        )
    ) == [1, 5, None]  # ATOW we get [1, None, None]

    # Fails ATOW. Same as above the Y argument is not even an optional type (but X still is).
    assert to_list(
        ak.where(
            ak.Array([True, False, None]),
            ak.Array([1, None, None]),
            ak.Array([4, 5, 6]),
        )
    ) == [1, 5, None]  # ATOW we get [1, None, None]


def test_ak_where_with_optionals_multidim():
    # This needs to continue to work:
    assert to_list(
        ak.where(
            ak.Array([True, False]),
            ak.Array([[1, 2], [3, 4]]),
            ak.Array([[10, 11], [12, 13]]),
        )
    ) == [[1, 2], [12, 13]]

    # Option types only in X, not condition
    assert to_list(
        ak.where(
            ak.Array([[True, True], [False, False]]),
            ak.Array([[1, 2], None]),
            ak.Array([[10, 11], [12, 13]]),
        )
    ) == [[1, 2], [12, 13]]

    # Option types in condition and X, only one level of depth
    assert to_list(
        ak.where(
            ak.Array([[True, True], [False, False], [True, False], None]),
            ak.Array([[1, 2], None, None, [7, 8]]),
            ak.Array([[11, 12], [13, 14], [15, 16], [17, 18]]),
        )
    ) == [[1, 2], [13, 14], [None, 16], [None, None]]
    # Note: [[1, 2], [13, 14], [None, 16], None] might seem more natural,
    #       but broadcasting expands these arrays out.

    assert to_list(
        ak.where(
            ak.Array([[True, False], [True, None]]),
            ak.Array([1, 2]),
            ak.Array([None, 12]),
        )
    ) == [[1, None], [2, None]]


def test_ak_where_more_option_types():
    assert to_list(
        ak.where(
            ak.Array([False, True, None]),
            ak.Array(["this", None, "that"]),
            ak.Array(["foo", "bar", "baz"]),
        )
    ) == ["foo", None, None]

    bitmasked5 = ak.contents.BitMaskedArray(
        mask=ak.index.Index(
            np.array(
                [
                    0b10100,
                ],
                dtype=np.uint8,
            )
        ),
        content=ak.contents.NumpyArray(np.arange(5)),
        valid_when=False,
        length=5,
        lsb_order=True,
        parameters={"_my_param": "boysenberry"},
    )  # [0, 1, None, 3, None]
    unmasked5 = ak.contents.UnmaskedArray(
        ak.contents.NumpyArray(np.arange(10, 15))
    )  # [10, 11, 12, 13, 14]
    union5 = ak.Array([True, None, "two", 3, 4.4])

    mixed_result = ak.where(
        ak.Array([True, None, True, False, True]), bitmasked5, unmasked5
    )
    assert to_list(mixed_result) == [0, None, None, 13, None]
    assert (
        mixed_result.layout.parameters.get("_my_param") is None
    )  # Params not preserved.

    assert to_list(
        ak.where(ak.Array([True, True, True, False, None]), union5, unmasked5)
    ) == [True, None, "two", 13, None]



from __future__ import annotations

import cupy as cp
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


primes = [x for x in range(2, 1000) if all(x % n != 0 for n in range(2, x))]


def test_0115_generic_reducer_operation_ListOffsetArray_to_RegularArray():
    content = ak.contents.NumpyArray(np.array(primes[: 2 * 3 * 5], dtype=np.int64))
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, content)
    regulararray = listoffsetarray.to_RegularArray()
    cuda_listoffsetarray = ak.to_backend(listoffsetarray, "cuda")
    cuda_regulararray = ak.to_backend(regulararray, "cuda")

    assert to_list(cuda_listoffsetarray) == to_list(cuda_regulararray)
    del cuda_listoffsetarray, cuda_regulararray


def test_0115_generic_reducer_operation_dimension_optiontype_1():
    content = ak.contents.NumpyArray(np.array(primes[: 2 * 3 * 5], dtype=np.int64))
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, content)
    index = ak.index.Index64(np.array([5, -1, 3, 2, -1, 0], dtype=np.int64))
    indexedarray = ak.contents.IndexedOptionArray(index, listoffsetarray)
    depth2 = ak.contents.RegularArray(indexedarray, 3)
    depth2 = ak.to_backend(depth2, "cuda")

    assert to_list(depth2) == [
        [[101, 103, 107, 109, 113], None, [53, 59, 61, 67, 71]],
        [[31, 37, 41, 43, 47], None, [2, 3, 5, 7, 11]],
    ]
    assert to_list(ak.prod(depth2, axis=-1, keepdims=False, highlevel=False)) == [
        [101 * 103 * 107 * 109 * 113, None, 53 * 59 * 61 * 67 * 71],
        [31 * 37 * 41 * 43 * 47, None, 2 * 3 * 5 * 7 * 11],
    ]
    assert to_list(ak.prod(depth2, axis=-1, keepdims=True, highlevel=False)) == [
        [[101 * 103 * 107 * 109 * 113], None, [53 * 59 * 61 * 67 * 71]],
        [[31 * 37 * 41 * 43 * 47], None, [2 * 3 * 5 * 7 * 11]],
    ]
    del depth2


def test_0115_generic_reducer_operation_dimension_optiontype_2():
    content = ak.contents.NumpyArray(np.array(primes[: 2 * 3 * 5], dtype=np.int64))
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, content)
    index = ak.index.Index64(np.array([5, 4, 3, 2, 1, 0], dtype=np.int64))
    indexedarray = ak.contents.IndexedArray(index, listoffsetarray)
    depth2 = ak.contents.RegularArray(indexedarray, 3)
    depth2 = ak.to_backend(depth2, "cuda")

    assert to_list(depth2) == [
        [[101, 103, 107, 109, 113], [73, 79, 83, 89, 97], [53, 59, 61, 67, 71]],
        [[31, 37, 41, 43, 47], [13, 17, 19, 23, 29], [2, 3, 5, 7, 11]],
    ]
    assert to_list(ak.prod(depth2, axis=-1, highlevel=False)) == [
        [101 * 103 * 107 * 109 * 113, 73 * 79 * 83 * 89 * 97, 53 * 59 * 61 * 67 * 71],
        [31 * 37 * 41 * 43 * 47, 13 * 17 * 19 * 23 * 29, 2 * 3 * 5 * 7 * 11],
    ]
    assert to_list(ak.prod(depth2, axis=-1, keepdims=True, highlevel=False)) == [
        [
            [101 * 103 * 107 * 109 * 113],
            [73 * 79 * 83 * 89 * 97],
            [53 * 59 * 61 * 67 * 71],
        ],
        [[31 * 37 * 41 * 43 * 47], [13 * 17 * 19 * 23 * 29], [2 * 3 * 5 * 7 * 11]],
    ]
    del depth2


def test_0115_generic_reducer_operation_reproduce_numpy_1():
    content1 = ak.contents.NumpyArray(np.array(primes[: 2 * 3 * 5], dtype=np.int64))
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(
        offsets2, ak.contents.ListOffsetArray(offsets1, content1)
    )
    depth2 = ak.to_backend(depth2, "cuda", highlevel=False)

    assert to_list(depth2) == [
        [[2, 3, 5, 7, 11], [13, 17, 19, 23, 29], [31, 37, 41, 43, 47]],
        [[53, 59, 61, 67, 71], [73, 79, 83, 89, 97], [101, 103, 107, 109, 113]],
    ]

    assert to_list(ak.prod(depth2, axis=-1, highlevel=False)) == [
        [2 * 3 * 5 * 7 * 11, 13 * 17 * 19 * 23 * 29, 31 * 37 * 41 * 43 * 47],
        [53 * 59 * 61 * 67 * 71, 73 * 79 * 83 * 89 * 97, 101 * 103 * 107 * 109 * 113],
    ]
    assert (
        ak.prod(depth2.to_typetracer(), axis=-1, highlevel=False).form
        == ak.prod(depth2, axis=-1, highlevel=False).form
    )
    assert to_list(ak.prod(depth2, axis=2, highlevel=False)) == [
        [2 * 3 * 5 * 7 * 11, 13 * 17 * 19 * 23 * 29, 31 * 37 * 41 * 43 * 47],
        [53 * 59 * 61 * 67 * 71, 73 * 79 * 83 * 89 * 97, 101 * 103 * 107 * 109 * 113],
    ]
    assert (
        ak.prod(depth2.to_typetracer(), axis=2, highlevel=False).form
        == ak.prod(depth2, axis=2, highlevel=False).form
    )
    del depth2


def test_0115_generic_reducer_operation_reproduce_numpy_2():
    content2 = ak.contents.NumpyArray(np.array(primes[:12], dtype=np.int64))
    offsets3 = ak.index.Index64(np.array([0, 4, 8, 12], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(ak.prod(depth1, -1, highlevel=False)) == [
        2 * 3 * 5 * 7,
        11 * 13 * 17 * 19,
        23 * 29 * 31 * 37,
    ]
    assert (
        ak.prod(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.prod(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.prod(depth1, 1, highlevel=False)) == [
        2 * 3 * 5 * 7,
        11 * 13 * 17 * 19,
        23 * 29 * 31 * 37,
    ]
    assert (
        ak.prod(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.prod(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_gaps_1():
    content1 = ak.contents.NumpyArray(np.array(primes[: 2 * 3 * 5], dtype=np.int64))
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(
        offsets2, ak.contents.ListOffsetArray(offsets1, content1)
    )
    depth2 = ak.to_backend(depth2, "cuda", highlevel=False)

    assert to_list(depth2) == [
        [[2, 3, 5, 7, 11], [13, 17, 19, 23, 29], [31, 37, 41, 43, 47]],
        [[53, 59, 61, 67, 71], [73, 79, 83, 89, 97], [101, 103, 107, 109, 113]],
    ]

    assert to_list(ak.prod(depth2, -1, highlevel=False)) == [
        [2 * 3 * 5 * 7 * 11, 13 * 17 * 19 * 23 * 29, 31 * 37 * 41 * 43 * 47],
        [53 * 59 * 61 * 67 * 71, 73 * 79 * 83 * 89 * 97, 101 * 103 * 107 * 109 * 113],
    ]
    assert (
        ak.prod(depth2.to_typetracer(), -1, highlevel=False).form
        == ak.prod(depth2, -1, highlevel=False).form
    )

    del depth2


def test_0115_generic_reducer_operation_gaps_2():
    content1 = ak.contents.NumpyArray(np.array(primes[:9], dtype=np.int64))
    offsets1 = ak.index.Index64(np.array([0, 3, 3, 5, 6, 8, 9], dtype=np.int64))
    offsets2 = ak.index.Index64(np.array([0, 4, 4, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(
        offsets2, ak.contents.ListOffsetArray(offsets1, content1)
    )
    depth2 = ak.to_backend(depth2, "cuda", highlevel=False)

    assert to_list(depth2) == [
        [[2, 3, 5], [], [7, 11], [13]],
        [],
        [[17, 19], [23]],
    ]

    assert to_list(ak.prod(depth2, -1, highlevel=False)) == [
        [2 * 3 * 5, 1, 7 * 11, 13],
        [],
        [17 * 19, 23],
    ]
    assert (
        ak.prod(depth2.to_typetracer(), -1, highlevel=False).form
        == ak.prod(depth2, -1, highlevel=False).form
    )

    del depth2


def test_0115_generic_reducer_operation_complicated():
    offsets1 = ak.index.Index64(np.array([0, 3, 3, 5], dtype=np.int64))
    content1 = ak.contents.ListOffsetArray(
        offsets1, ak.contents.NumpyArray(np.array(primes[:5], dtype=np.int64))
    )
    offsets2 = ak.index.Index64(np.array([0, 3, 3, 5, 6, 8, 9], dtype=np.int64))
    offsets3 = ak.index.Index64(np.array([0, 4, 4, 6], dtype=np.int64))
    content2 = ak.contents.ListOffsetArray(
        offsets3,
        ak.contents.ListOffsetArray(
            offsets2, ak.contents.NumpyArray(np.array(primes[:9], dtype=np.int64))
        ),
    )
    offsets4 = ak.index.Index64(np.array([0, 1, 1, 3], dtype=np.int64))
    complicated = ak.contents.ListOffsetArray(
        offsets4, ak.contents.RecordArray([content1, content2], ["x", "y"])
    )
    complicated = ak.to_backend(complicated, "cuda", highlevel=False)

    assert to_list(complicated) == [
        [{"x": [2, 3, 5], "y": [[2, 3, 5], [], [7, 11], [13]]}],
        [],
        [{"x": [], "y": []}, {"x": [7, 11], "y": [[17, 19], [23]]}],
    ]

    assert to_list(complicated["x"]) == [[[2, 3, 5]], [], [[], [7, 11]]]
    assert complicated.to_typetracer()["x"].form == complicated["x"].form
    assert to_list(complicated["y"]) == [
        [[[2, 3, 5], [], [7, 11], [13]]],
        [],
        [[], [[17, 19], [23]]],
    ]
    assert complicated.to_typetracer()["y"].form == complicated["y"].form

    with pytest.raises(TypeError):
        to_list(ak.prod(complicated, -1, highlevel=False))

    with pytest.raises(TypeError):
        assert (
            ak.prod(complicated.to_typetracer(), -1, highlevel=False).form
            == ak.prod(complicated, -1, highlevel=False).form
        )

    assert to_list(ak.prod(complicated["x"], -1, highlevel=False)) == [
        [30],
        [],
        [1, 77],
    ]
    assert (
        ak.prod(complicated.to_typetracer()["x"], -1, highlevel=False).form
        == ak.prod(complicated["x"], -1, highlevel=False).form
    )
    assert to_list(ak.prod(complicated["y"], -1, highlevel=False)) == [
        [[30, 1, 77, 13]],
        [],
        [[], [323, 23]],
    ]
    assert (
        ak.prod(complicated.to_typetracer()["y"], -1, highlevel=False).form
        == ak.prod(complicated["y"], -1, highlevel=False).form
    )

    with pytest.raises(TypeError):
        to_list(ak.prod(complicated[0], -1, highlevel=False))

    with pytest.raises(TypeError):
        to_list(ak.prod(complicated.to_typetracer()[0], -1, highlevel=False))
    del complicated


def test_0115_generic_reducer_operation_EmptyArray():
    offsets = ak.index.Index64(np.array([0, 0, 0, 0], dtype=np.int64))
    array = ak.contents.ListOffsetArray(offsets, ak.contents.EmptyArray())
    array = ak.to_backend(array, "cuda", highlevel=False)
    assert to_list(array) == [[], [], []]

    assert to_list(ak.prod(array, -1, highlevel=False)) == [1, 1, 1]
    assert (
        ak.prod(array.to_typetracer(), -1, highlevel=False).form
        == ak.prod(array, -1, highlevel=False).form
    )

    offsets = ak.index.Index64(np.array([0, 0, 0, 0], dtype=np.int64))
    array = ak.contents.ListOffsetArray(
        offsets, ak.contents.NumpyArray(np.array([], dtype=np.int64))
    )
    array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(array) == [[], [], []]

    assert to_list(ak.prod(array, -1, highlevel=False)) == [1, 1, 1]
    assert (
        ak.prod(array.to_typetracer(), -1, highlevel=False).form
        == ak.prod(array, -1, highlevel=False).form
    )
    del array


def test_0115_generic_reducer_operation_IndexedOptionArray_1():
    content = ak.contents.NumpyArray(np.array(primes[: 2 * 3 * 5], dtype=np.int64))
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, content)
    index = ak.index.Index64(np.array([5, 4, 3, 2, 1, 0], dtype=np.int64))
    indexedarray = ak.contents.IndexedArray(index, listoffsetarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(offsets2, indexedarray)
    depth2 = ak.to_backend(depth2, "cuda", highlevel=False)

    assert to_list(depth2) == [
        [[101, 103, 107, 109, 113], [73, 79, 83, 89, 97], [53, 59, 61, 67, 71]],
        [[31, 37, 41, 43, 47], [13, 17, 19, 23, 29], [2, 3, 5, 7, 11]],
    ]

    assert to_list(ak.prod(depth2, -1, highlevel=False)) == [
        [101 * 103 * 107 * 109 * 113, 73 * 79 * 83 * 89 * 97, 53 * 59 * 61 * 67 * 71],
        [31 * 37 * 41 * 43 * 47, 13 * 17 * 19 * 23 * 29, 2 * 3 * 5 * 7 * 11],
    ]
    assert (
        ak.prod(depth2.to_typetracer(), -1, highlevel=False).form
        == ak.prod(depth2, -1, highlevel=False).form
    )

    del depth2


def test_0115_generic_reducer_operation_IndexedOptionArray_2():
    content = ak.contents.NumpyArray(
        np.array(
            [
                2,
                3,
                5,
                7,
                11,
                31,
                37,
                41,
                43,
                47,
                53,
                59,
                61,
                67,
                71,
                101,
                103,
                107,
                109,
                113,
            ],
            dtype=np.int64,
        )
    )
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, content)
    index = ak.index.Index64(np.array([3, -1, 2, 1, -1, 0], dtype=np.int64))
    indexedoptionarray = ak.contents.IndexedOptionArray(index, listoffsetarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(offsets2, indexedoptionarray)
    depth2 = ak.to_backend(depth2, "cuda", highlevel=False)

    assert to_list(depth2) == [
        [[101, 103, 107, 109, 113], None, [53, 59, 61, 67, 71]],
        [[31, 37, 41, 43, 47], None, [2, 3, 5, 7, 11]],
    ]

    assert to_list(ak.prod(depth2, -1, highlevel=False)) == [
        [101 * 103 * 107 * 109 * 113, None, 53 * 59 * 61 * 67 * 71],
        [31 * 37 * 41 * 43 * 47, None, 2 * 3 * 5 * 7 * 11],
    ]
    assert (
        ak.prod(depth2.to_typetracer(), -1, highlevel=False).form
        == ak.prod(depth2, -1, highlevel=False).form
    )

    del depth2


def test_0115_generic_reducer_operation_IndexedOptionArray_3():
    content = ak.contents.NumpyArray(
        np.array(
            [
                2,
                3,
                5,
                7,
                11,
                31,
                37,
                41,
                43,
                47,
                53,
                59,
                61,
                67,
                71,
                101,
                103,
                107,
                109,
                113,
            ],
            dtype=np.int64,
        )
    )
    index = ak.index.Index64(
        np.array(
            [
                15,
                16,
                17,
                18,
                19,
                -1,
                -1,
                -1,
                -1,
                -1,
                10,
                11,
                12,
                13,
                14,
                5,
                6,
                7,
                8,
                9,
                -1,
                -1,
                -1,
                -1,
                -1,
                0,
                1,
                2,
                3,
                4,
            ],
            dtype=np.int64,
        )
    )
    indexedoptionarray = ak.contents.IndexedOptionArray(index, content)
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, indexedoptionarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(offsets2, listoffsetarray)
    depth2 = ak.to_backend(depth2, "cuda", highlevel=False)

    assert to_list(depth2) == [
        [
            [101, 103, 107, 109, 113],
            [None, None, None, None, None],
            [53, 59, 61, 67, 71],
        ],
        [[31, 37, 41, 43, 47], [None, None, None, None, None], [2, 3, 5, 7, 11]],
    ]

    assert to_list(ak.prod(depth2, -1, highlevel=False)) == [
        [101 * 103 * 107 * 109 * 113, 1 * 1 * 1 * 1 * 1, 53 * 59 * 61 * 67 * 71],
        [31 * 37 * 41 * 43 * 47, 1 * 1 * 1 * 1 * 1, 2 * 3 * 5 * 7 * 11],
    ]
    assert (
        ak.prod(depth2.to_typetracer(), -1, highlevel=False).form
        == ak.prod(depth2, -1, highlevel=False).form
    )

    del depth2


def test_0115_generic_reducer_operation_IndexedOptionArray_4():
    content = ak.contents.NumpyArray(
        np.array(
            [
                2,
                3,
                5,
                7,
                11,
                31,
                37,
                41,
                43,
                47,
                53,
                59,
                61,
                67,
                71,
                101,
                103,
                107,
                109,
                113,
            ],
            dtype=np.int64,
        )
    )
    index = ak.index.Index64(
        np.array(
            [
                15,
                16,
                17,
                18,
                19,
                -1,
                10,
                11,
                12,
                13,
                14,
                5,
                6,
                7,
                8,
                9,
                -1,
                0,
                1,
                2,
                3,
                4,
            ],
            dtype=np.int64,
        )
    )
    indexedoptionarray = ak.contents.IndexedOptionArray(index, content)
    offsets1 = ak.index.Index64(np.array([0, 5, 6, 11, 16, 17, 22], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, indexedoptionarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(offsets2, listoffsetarray)
    depth2 = ak.to_backend(depth2, "cuda", highlevel=False)

    assert to_list(depth2) == [
        [[101, 103, 107, 109, 113], [None], [53, 59, 61, 67, 71]],
        [[31, 37, 41, 43, 47], [None], [2, 3, 5, 7, 11]],
    ]

    assert to_list(ak.prod(depth2, -1, highlevel=False)) == [
        [101 * 103 * 107 * 109 * 113, 1, 53 * 59 * 61 * 67 * 71],
        [31 * 37 * 41 * 43 * 47, 1, 2 * 3 * 5 * 7 * 11],
    ]
    assert (
        ak.prod(depth2.to_typetracer(), -1, highlevel=False).form
        == ak.prod(depth2, -1, highlevel=False).form
    )

    del depth2


def test_0115_generic_reducer_operation_sum():
    content2 = ak.contents.NumpyArray(
        np.array([1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048], dtype=np.int64)
    )
    offsets3 = ak.index.Index64(np.array([0, 4, 8, 12], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(ak.sum(depth1, -1, highlevel=False)) == [
        1 + 2 + 4 + 8,
        16 + 32 + 64 + 128,
        256 + 512 + 1024 + 2048,
    ]
    assert (
        ak.sum(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.sum(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.sum(depth1, 1, highlevel=False)) == [
        1 + 2 + 4 + 8,
        16 + 32 + 64 + 128,
        256 + 512 + 1024 + 2048,
    ]
    assert (
        ak.sum(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.sum(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_any():
    content2 = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 0.0, 0.0, 0.0])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 0.0, 0.0, 0.0],
    ]

    assert to_list(ak.any(depth1, -1, highlevel=False)) == [True, True, False]
    assert (
        ak.any(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.any(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.any(depth1, 1, highlevel=False)) == [True, True, False]
    assert (
        ak.any(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.any(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_all():
    content2 = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4],
    ]

    assert to_list(ak.all(depth1, -1, highlevel=False)) == [True, False, False]
    assert (
        ak.all(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.all(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.all(depth1, 1, highlevel=False)) == [True, False, False]
    assert (
        ak.all(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.all(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_count():
    content2 = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4],
    ]

    assert to_list(ak.count(depth1, -1, highlevel=False)) == [3, 3, 4]
    assert (
        ak.count(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.count(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.count(depth1, 1, highlevel=False)) == [3, 3, 4]
    assert (
        ak.count(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.count(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_count_nonzero():
    content2 = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4],
    ]

    assert to_list(ak.count_nonzero(depth1, -1, highlevel=False)) == [3, 1, 2]
    assert (
        ak.count_nonzero(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.count_nonzero(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.count_nonzero(depth1, 1, highlevel=False)) == [3, 1, 2]
    assert (
        ak.count_nonzero(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.count_nonzero(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_count_min_1():
    content2 = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4],
    ]

    assert to_list(ak.min(depth1, -1, highlevel=False)) == [1.1, 0.0, 0.0]
    assert (
        ak.min(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.min(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.min(depth1, 1, highlevel=False)) == [1.1, 0.0, 0.0]
    assert (
        ak.min(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.min(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_count_min_2():
    content2 = ak.contents.NumpyArray(
        np.array([True, True, True, False, True, False, False, True, False, True])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [True, True, True],
        [False, True, False],
        [False, True, False, True],
    ]

    assert to_list(ak.min(depth1, -1, highlevel=False)) == [True, False, False]
    assert (
        ak.min(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.min(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.min(depth1, 1, highlevel=False)) == [True, False, False]
    assert (
        ak.min(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.min(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_count_max_1():
    content2 = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [1.1, 2.2, 3.3],
        [0.0, 2.2, 0.0],
        [0.0, 2.2, 0.0, 4.4],
    ]

    assert to_list(ak.max(depth1, -1, highlevel=False)) == [3.3, 2.2, 4.4]
    assert (
        ak.max(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.max(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.max(depth1, 1, highlevel=False)) == [3.3, 2.2, 4.4]
    assert (
        ak.max(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.max(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_count_max_2():
    content2 = ak.contents.NumpyArray(
        np.array([False, True, True, False, True, False, False, False, False, False])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [False, True, True],
        [False, True, False],
        [False, False, False, False],
    ]

    assert to_list(ak.max(depth1, -1, highlevel=False)) == [True, True, False]
    assert (
        ak.max(depth1.to_typetracer(), -1, highlevel=False).form
        == ak.max(depth1, -1, highlevel=False).form
    )
    assert to_list(ak.max(depth1, 1, highlevel=False)) == [True, True, False]
    assert (
        ak.max(depth1.to_typetracer(), 1, highlevel=False).form
        == ak.max(depth1, 1, highlevel=False).form
    )

    del depth1


def test_0115_generic_reducer_operation_mask():
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 6, 6, 9], dtype=np.int64))
    array = ak.contents.ListOffsetArray(offsets, content)
    array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(ak.min(array, axis=-1, mask_identity=False, highlevel=False)) == [
        1.1,
        np.inf,
        4.4,
        6.6,
        np.inf,
        np.inf,
        7.7,
    ]
    assert (
        ak.min(
            array.to_typetracer(), axis=-1, mask_identity=False, highlevel=False
        ).form
        == ak.min(array, axis=-1, mask_identity=False, highlevel=False).form
    )
    assert to_list(ak.min(array, axis=-1, mask_identity=True, highlevel=False)) == [
        1.1,
        None,
        4.4,
        6.6,
        None,
        None,
        7.7,
    ]
    assert (
        ak.min(array.to_typetracer(), axis=-1, mask_identity=True, highlevel=False).form
        == ak.min(array, axis=-1, mask_identity=True, highlevel=False).form
    )
    del array


def test_0115_generic_reducer_operation_ByteMaskedArray():
    content = ak.operations.from_iter(
        [
            [[1.1, 0.0, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 9.9, 8.8, 7.7]],
            [[], [12.2, 11.1, 10.0]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    v2_array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    v2_array = ak.to_backend(v2_array, "cuda", highlevel=False)

    assert to_list(v2_array) == [
        [[1.1, 0.0, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [12.2, 11.1, 10.0]],
    ]
    assert to_list(ak.argmin(v2_array, axis=-1, highlevel=False)) == [
        [1, None, 0],
        [],
        None,
        None,
        [None, 2],
    ]
    assert (
        ak.argmin(v2_array.to_typetracer(), axis=-1, highlevel=False).form
        == ak.argmin(v2_array, axis=-1, highlevel=False).form
    )
    del v2_array


def test_0115_generic_reducer_operation_keepdims():
    nparray = np.array(primes[: 2 * 3 * 5], dtype=np.int64).reshape(2, 3, 5)
    content1 = ak.contents.NumpyArray(np.array(primes[: 2 * 3 * 5], dtype=np.int64))
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(
        offsets2, ak.contents.ListOffsetArray(offsets1, content1)
    )
    depth2 = ak.to_backend(depth2, "cuda", highlevel=False)

    assert to_list(depth2) == [
        [[2, 3, 5, 7, 11], [13, 17, 19, 23, 29], [31, 37, 41, 43, 47]],
        [[53, 59, 61, 67, 71], [73, 79, 83, 89, 97], [101, 103, 107, 109, 113]],
    ]

    assert to_list(
        ak.prod(depth2, axis=-1, keepdims=False, highlevel=False)
    ) == to_list(ak.prod(nparray, axis=-1, keepdims=False, highlevel=False))
    assert (
        ak.prod(depth2.to_typetracer(), axis=-1, keepdims=False, highlevel=False).form
        == ak.prod(depth2, axis=-1, keepdims=False, highlevel=False).form
    )

    assert to_list(ak.prod(depth2, axis=-1, keepdims=True, highlevel=False)) == to_list(
        ak.prod(nparray, axis=-1, keepdims=True, highlevel=False)
    )
    assert (
        ak.prod(depth2.to_typetracer(), axis=-1, keepdims=True, highlevel=False).form
        == ak.prod(depth2, axis=-1, keepdims=True, highlevel=False).form
    )

    del depth2


def test_0115_generic_reducer_operation_highlevel_1():
    array = ak.highlevel.Array(
        [[[2, 3, 5], [], [7, 11], [13]], [], [[17, 19], [23]]], check_valid=True
    )
    array = ak.to_backend(array, "cuda", highlevel=False)

    assert ak.operations.count(array) == 9
    assert to_list(ak.operations.count(array, axis=-1)) == [
        [3, 0, 2, 1],
        [],
        [2, 1],
    ]
    assert to_list(ak.operations.count(array, axis=-1, keepdims=True)) == [
        [[3], [0], [2], [1]],
        [],
        [[2], [1]],
    ]

    assert ak.operations.count_nonzero(array) == 9
    assert to_list(ak.operations.count_nonzero(array, axis=-1)) == [
        [3, 0, 2, 1],
        [],
        [2, 1],
    ]

    assert ak.operations.sum(array) == 2 + 3 + 5 + 7 + 11 + 13 + 17 + 19 + 23
    assert to_list(ak.operations.sum(array, axis=-1)) == [
        [2 + 3 + 5, 0, 7 + 11, 13],
        [],
        [17 + 19, 23],
    ]

    assert ak.operations.prod(array) == 2 * 3 * 5 * 7 * 11 * 13 * 17 * 19 * 23
    assert to_list(ak.operations.prod(array, axis=-1)) == [
        [2 * 3 * 5, 1, 7 * 11, 13],
        [],
        [17 * 19, 23],
    ]

    assert ak.operations.max(array) == 23
    assert to_list(ak.operations.max(array, axis=-1)) == [
        [5, None, 11, 13],
        [],
        [19, 23],
    ]

    del array


def test_0115_generic_reducer_operation_highlevel_2():
    array = ak.highlevel.Array(
        [
            [[True, False, True], [], [False, False], [True]],
            [],
            [[False, True], [True]],
        ],
        check_valid=True,
    )
    array = ak.to_backend(array, "cuda")
    assert ak.operations.any(array) == cp.bool_(True)
    assert to_list(ak.operations.any(array, axis=-1)) == [
        [True, False, False, True],
        [],
        [True, True],
    ]

    assert ak.operations.all(array) == cp.bool_(False)
    assert to_list(ak.operations.all(array, axis=-1)) == [
        [False, True, False, True],
        [],
        [False, True],
    ]
    del array


def test_nonreducers():
    x = ak.highlevel.Array([[1, 2, 3, 4, 5], [1, 2, 3, 4, 5]], check_valid=True)
    y = ak.highlevel.Array(
        [[1.1, 2.2, 2.9, 4.0, 5.1], [0.9, 2.1, 3.2, 4.1, 4.9]], check_valid=True
    )
    x = ak.to_backend(x, "cuda")
    y = ak.to_backend(y, "cuda")

    cpt.assert_allclose(ak.operations.mean(y), cp.mean(ak.operations.to_numpy(y)))
    cpt.assert_allclose(ak.operations.var(y), cp.var(ak.operations.to_numpy(y)))
    cpt.assert_allclose(
        ak.operations.var(y, ddof=1), cp.var(ak.operations.to_numpy(y), ddof=1)
    )
    cpt.assert_allclose(ak.operations.std(y), np.std(ak.operations.to_numpy(y)))
    cpt.assert_allclose(
        ak.operations.std(y, ddof=1), cp.std(ak.operations.to_numpy(y), ddof=1)
    )

    cpt.assert_allclose(ak.operations.moment(y, 1), cp.mean(ak.operations.to_numpy(y)))
    cpt.assert_allclose(
        ak.operations.moment(y - ak.operations.mean(y), 2),
        cp.var(ak.operations.to_numpy(y)),
    )
    cpt.assert_allclose(ak.operations.covar(y, y), cp.var(ak.operations.to_numpy(y)))
    cpt.assert_allclose(ak.operations.corr(y, y), 1.0)

    cpt.assert_allclose(ak.operations.corr(x, y), 0.9968772535047296)

    cpt.assert_allclose(
        to_list(ak.operations.mean(y, axis=-1)),
        to_list(cp.mean(ak.operations.to_numpy(y), axis=-1)),
    )
    cpt.assert_allclose(
        to_list(ak.operations.var(y, axis=-1)),
        to_list(cp.var(ak.operations.to_numpy(y), axis=-1)),
    )
    cpt.assert_allclose(
        to_list(ak.operations.var(y, axis=-1, ddof=1)),
        to_list(cp.var(ak.operations.to_numpy(y), axis=-1, ddof=1)),
    )
    cpt.assert_allclose(
        to_list(ak.operations.std(y, axis=-1)),
        to_list(cp.std(ak.operations.to_numpy(y), axis=-1)),
    )
    cpt.assert_allclose(
        to_list(ak.operations.std(y, axis=-1, ddof=1)),
        to_list(cp.std(ak.operations.to_numpy(y), axis=-1, ddof=1)),
    )

    cpt.assert_allclose(
        to_list(ak.operations.moment(y, 1, axis=-1)),
        to_list(cp.mean(ak.operations.to_numpy(y), axis=-1)),
    )
    cpt.assert_allclose(
        to_list(ak.operations.moment(y - ak.operations.mean(y, axis=-1), 2, axis=-1)),
        to_list(cp.var(ak.operations.to_numpy(y), axis=-1)),
    )
    cpt.assert_allclose(
        to_list(ak.operations.covar(y, y, axis=-1)),
        to_list(cp.var(ak.operations.to_numpy(y), axis=-1)),
    )
    cpt.assert_allclose(to_list(ak.operations.corr(y, y, axis=-1)), [1.0, 1.0])

    cpt.assert_allclose(
        to_list(ak.operations.corr(x, y, axis=-1)),
        [0.9975103695813371, 0.9964193240901015],
    )


def test_softmax():
    array = ak.highlevel.Array(
        [[np.log(2), np.log(2), np.log(4)], [], [np.log(5), np.log(5)]],
        check_valid=True,
    )
    array = ak.to_backend(array, "cuda")

    assert to_list(ak.operations.softmax(array, axis=-1)) == [
        pytest.approx([0.25, 0.25, 0.5]),
        [],
        pytest.approx([0.5, 0.5]),
    ]
    del array


def test_prod_bool_1():
    # this had been silently broken
    array = np.array([[True, False, False], [True, False, False]])
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda")

    assert to_list(ak.prod(depth1, axis=-1, highlevel=False)) == [0, 1, 0, 0]
    assert to_list(ak.all(depth1, axis=-1, highlevel=False)) == [
        False,
        True,
        False,
        False,
    ]
    assert to_list(ak.min(depth1, axis=-1, highlevel=False)) == [
        False,
        None,
        False,
        False,
    ]
    del depth1


def test_prod_bool_2():
    array = np.array([[True, False, False], [True, False, False]]).view(np.uint8)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda")

    assert to_list(ak.prod(depth1, axis=-1, highlevel=False)) == [0, 1, 0, 0]
    assert to_list(ak.all(depth1, axis=-1, highlevel=False)) == [0, 1, 0, 0]
    assert to_list(ak.min(depth1, axis=-1, highlevel=False)) == [0, None, 0, 0]

    array = np.array([[True, False, False], [True, False, False]]).astype(np.int32)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda")

    assert to_list(ak.prod(depth1, axis=-1, highlevel=False)) == [0, 1, 0, 0]
    assert to_list(ak.all(depth1, axis=-1, highlevel=False)) == [0, 1, 0, 0]
    assert to_list(ak.min(depth1, axis=-1, highlevel=False)) == [0, None, 0, 0]
    del depth1



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

# from __future__ import annotations
from __future__ import annotations

import io

import numpy as np
import pytest

import awkward as ak
from awkward.contents import (
    EmptyArray,
    NumpyArray,
    RecordArray,
    RegularArray,
    UnionArray,
)
from awkward.operations import to_list
from awkward.types import ListType, OptionType, UnknownType

pyarrow = pytest.importorskip("pyarrow")
pq = pytest.importorskip("pyarrow.parquet")


def test_bare_unknown():
    unk_array = ak.Array([[], []])  # sub-arrays are type unknown
    assert unk_array[0].type.content == UnknownType()

    unk_array_arrow = ak.to_arrow(unk_array)
    # This array has one field, where we can get at pyarrow storage info
    if hasattr(unk_array_arrow.type.storage_type, "field"):
        # We support older versions of pyarrow that lack a field method
        field = unk_array_arrow.type.storage_type.field(0)
        assert field.type.storage_type == pyarrow.null()
        assert field.nullable  # Must be nullable to be valid in Arrow
    array_is_valid_within_parquet(unk_array_arrow)

    orig_array = ak.from_arrow(unk_array_arrow)
    assert to_list(orig_array) == [[], []]
    assert orig_array.type == unk_array.type


def test_option_unknown():
    unk_option_array = ak.Array([[None, None], []])  # type is ?unknown
    assert unk_option_array.type.content == ListType(OptionType(UnknownType()))

    unk_opt_array_arrow = ak.to_arrow(unk_option_array)
    if hasattr(unk_opt_array_arrow.type.storage_type, "field"):
        field = unk_opt_array_arrow.type.storage_type.field(0)
        assert field.type.storage_type == pyarrow.null()
        assert field.nullable  # Nullable but this time it's because we're ?unknown
    array_is_valid_within_parquet(unk_opt_array_arrow)

    orig_array = ak.from_arrow(unk_opt_array_arrow)
    assert orig_array.type == unk_option_array.type
    assert to_list(orig_array) == [[None, None], []]

    # This is different for ... reasons:
    ua1 = ak.Array([[], [], [], [None]])[0:3]
    assert ua1.type.content == ListType(OptionType(UnknownType()))
    ua1a = ak.to_arrow(ua1)
    if hasattr(ua1a.type.storage_type, "field"):
        field = ua1a.type.storage_type.field(0)
        assert field.type.storage_type == pyarrow.null()
        assert field.nullable  # Like above, still nullable
    ua1aa = ak.from_arrow(ua1a)
    assert ua1aa.type.content == ListType(OptionType(UnknownType()))
    assert len(ua1aa) == 3


def test_toplevel_unknown():
    unk_array = ak.Array([])
    assert unk_array.type.content == UnknownType()

    unk_array_arrow = ak.to_arrow(unk_array)
    assert len(unk_array_arrow) == 0
    # Note: we cannot test the nullability of this arrow array, since it has zero fields
    # field = unk_array_arrow.type.storage_type.field(0)
    assert unk_array_arrow.type.num_fields == 0  # (just demonstrating the above)
    # But we can still give Parquet conversion a shot
    array_is_valid_within_parquet(unk_array_arrow)

    orig_array = ak.from_arrow(unk_array_arrow)
    assert orig_array.type == unk_array.type
    assert to_list(orig_array) == []

    ua2 = ak.Array([None])[0:0]  # This is a top-level Option<EmptyArray>
    assert ua2.type.content == OptionType(UnknownType())
    ua2a = ak.to_arrow(ua2)
    ua2aa = ak.from_arrow(ua2a)
    assert ua2aa.type.content == OptionType(UnknownType())


def test_recordarray_with_unknowns():
    a = RecordArray([EmptyArray(), NumpyArray([])], ["x", "y"], length=0)
    arw = ak.to_arrow(a)
    if hasattr(arw.type.storage_type, "field"):
        assert arw.type.storage_type.field(0).nullable
    array_is_valid_within_parquet(arw)
    # This is a strange, laboratory kind of object.
    # It seems unlikely to be found in the wild.
    # I'm not sure what other tests here would be meaningful.


def test_table_with_unknowns():
    a = RecordArray([EmptyArray(), NumpyArray([1, 2])], ["x", "y"])
    # Again this is a strange one!
    table = ak.to_arrow_table(a)
    assert table.field(0).nullable
    temp = io.BytesIO()
    pq.write_table(table, temp)


def test_regulararray_with_unknown():
    a = RegularArray(EmptyArray(), 0)
    # RegularArray is helpful.
    # But when it's given an EmptyArray to recycle, it becomes absolutely unhelpful.
    arw = ak.to_arrow(a)
    if hasattr(arw.type.storage_type, "field"):
        assert arw.type.storage_type.field(0).nullable
    assert to_list(arw) == []
    array_is_valid_within_parquet(arw)


def test_unionarray_with_unknown():
    # Although a UnionArray with an EmptyArray content type has no application,
    # we can still exercise a code path this way.
    a = UnionArray(
        tags=ak.index.Index8(np.array([1, 1, 1], dtype=np.int8)),
        # tags: none of the elements are, or could be, taken from the EmptyArray.
        index=ak.index.Index64(np.array([0, 1, 2], dtype=np.int64)),
        contents=[
            EmptyArray(),
            NumpyArray([10, 20, 30]),
        ],
    )
    assert to_list(a) == [10, 20, 30]
    arw = ak.to_arrow(a)
    if hasattr(arw.type.storage_type, "field"):
        assert arw.type.storage_type.field(0).nullable
        assert not arw.type.storage_type.field(1).nullable
    # array_is_valid_within_parquet(arw)  # This fails for unrelated reasons.
    art = ak.from_arrow(arw)
    # round-trip is okay but the UnionArray is lost. Separate issue?
    assert to_list(art) == [10, 20, 30]


#### Helper method(s)


def array_is_valid_within_parquet(arrow_array):
    """
    Helper function that writes the given array to a Parquet table.
    Prior to 2340, this would raise:
    pyarrow.lib.ArrowInvalid: NullType Arrow field must be nullable
    """
    table = pyarrow.Table.from_arrays([arrow_array], names=["col1"])
    table.validate(
        full=True
    )  # Frustratingly, this does *not* produce the anticipated exception, but..
    temp = io.BytesIO()
    pq.write_table(table, temp)  # *this does* perform the validation we need.
    assert len(temp.getbuffer()) > 0



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
#
from __future__ import annotations

import pytest

import awkward as ak


def test():
    # taking the array directly from the issue: https://github.com/scikit-hep/awkward/issues/3619
    a = ak.Array([[1, 5], [3], [4]])

    # this is a valid case, where the index is not out of bounds
    c_valid = ak.Array([[1, 1], [0], [0]])
    assert a[c_valid].tolist() == [[5, 5], [3], [4]]

    # this is an invalid case, where the index for the second element `[1]` is out of bounds
    c_oob = ak.Array([[1, 1], [1], [0]])
    with pytest.raises(IndexError):
        a[c_oob]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward.operations import to_list


def test():
    onedim = ak.Array([1, 2, 3, None, 4, 5])
    result = ak.where(onedim)
    assert to_list(result[0]) == [0, 1, 2, 4, 5]
    assert to_list(result[0]) == ak.to_numpy(onedim).nonzero()[0].tolist()

    onedim = ak.Array([True, False, True, None, False])
    result = ak.where(onedim)
    assert to_list(result[0]) == [0, 2]
    assert to_list(result[0]) == ak.to_numpy(onedim).nonzero()[0].tolist()

    twodim = ak.Array([[1, 2, 3], [None, 4, 5]])
    result = ak.where(twodim)
    assert to_list(result[0]) == [0, 0, 0, 1, 1]
    assert to_list(result[0]) == ak.to_numpy(twodim).nonzero()[0].tolist()
    assert to_list(result[1]) == [0, 1, 2, 1, 2]
    assert to_list(result[1]) == ak.to_numpy(twodim).nonzero()[1].tolist()

    twodim = ak.Array([[True, False, True], [None, False, True]])
    result = ak.where(twodim)
    assert to_list(result[0]) == [0, 0, 1]
    assert to_list(result[0]) == ak.to_numpy(twodim).nonzero()[0].tolist()
    assert to_list(result[1]) == [0, 2, 2]
    assert to_list(result[1]) == ak.to_numpy(twodim).nonzero()[1].tolist()

    threedim = ak.Array([[[1, 2], [3, None]], [[4, 5], [None, None]]])
    result = ak.where(threedim)
    assert to_list(result[0]) == [0, 0, 0, 1, 1]
    assert to_list(result[0]) == ak.to_numpy(threedim).nonzero()[0].tolist()
    assert to_list(result[1]) == [0, 0, 1, 0, 0]
    assert to_list(result[1]) == ak.to_numpy(threedim).nonzero()[1].tolist()
    assert to_list(result[2]) == [0, 1, 0, 0, 1]
    assert to_list(result[2]) == ak.to_numpy(threedim).nonzero()[2].tolist()

    threedim = ak.Array([[[True, False], [True, None]], [[False, True], [None, None]]])
    result = ak.where(threedim)
    assert to_list(result[0]) == [0, 0, 1]
    assert to_list(result[0]) == ak.to_numpy(threedim).nonzero()[0].tolist()
    assert to_list(result[1]) == [0, 1, 0]
    assert to_list(result[1]) == ak.to_numpy(threedim).nonzero()[1].tolist()
    assert to_list(result[2]) == [0, 0, 1]
    assert to_list(result[2]) == ak.to_numpy(threedim).nonzero()[2].tolist()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy
import pytest

import awkward as ak


def test():
    behavior = {}

    class SuperVector:
        def add(self, other):
            """Add two vectors together elementwise using `x` and `y` components"""
            return ak.zip(
                {"x": self.x + other.x, "y": self.y + other.y},
                with_name="VectorTwoD",
                behavior=self.behavior,
            )

    # first sub-class
    @ak.mixin_class(behavior)
    class VectorTwoD(SuperVector):
        def __eq__(self, other):
            return ak.all(self.x == other.x) and ak.all(self.y == other.y)

    v = ak.Array(
        [
            [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}],
            [],
            [{"x": 3, "y": 3.3}],
            [
                {"x": 4, "y": 4.4},
                {"x": 5, "y": 5.5},
                {"x": 6, "y": 6.6},
            ],
        ],
        with_name="VectorTwoD",
        behavior=behavior,
    )
    v_added = ak.Array(
        [
            [{"x": 2, "y": 2.2}, {"x": 4, "y": 4.4}],
            [],
            [{"x": 6, "y": 6.6}],
            [
                {"x": 8, "y": 8.8},
                {"x": 10, "y": 11},
                {"x": 12, "y": 13.2},
            ],
        ],
        with_name="VectorTwoD",
        behavior=behavior,
    )

    # add method works but the binary operator does not
    assert v.add(v) == v_added
    with pytest.raises(TypeError):
        v + v

    # registering the operator makes everything work
    behavior[numpy.add, "VectorTwoD", "VectorTwoD"] = lambda v1, v2: v1.add(v2)
    assert v + v == v_added

    # instead of registering every operator again, just copy the behaviors of
    # another class to this class
    behavior.update(ak._util.copy_behaviors("VectorTwoD", "VectorTwoDAgain", behavior))

    # second sub-class
    @ak.mixin_class(behavior)
    class VectorTwoDAgain(VectorTwoD):
        pass

    v = ak.Array(
        [
            [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}],
            [],
            [{"x": 3, "y": 3.3}],
            [
                {"x": 4, "y": 4.4},
                {"x": 5, "y": 5.5},
                {"x": 6, "y": 6.6},
            ],
        ],
        with_name="VectorTwoDAgain",
        behavior=behavior,
    )
    assert v.add(v) == v_added
    assert v + v == v_added

    # instead of registering every operator again, just copy the behaviors of
    # another class to this class
    behavior.update(
        ak._util.copy_behaviors("VectorTwoDAgain", "VectorTwoDAgainAgain", behavior)
    )

    # third sub-class
    @ak.mixin_class(behavior)
    class VectorTwoDAgainAgain(VectorTwoDAgain):
        pass

    v = ak.Array(
        [
            [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}],
            [],
            [{"x": 3, "y": 3.3}],
            [
                {"x": 4, "y": 4.4},
                {"x": 5, "y": 5.5},
                {"x": 6, "y": 6.6},
            ],
        ],
        with_name="VectorTwoDAgainAgain",
        behavior=behavior,
    )
    assert v.add(v) == v_added
    assert v + v == v_added



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak


def test_only_floats_validation():
    class OnlyFloatsError(Exception): ...

    class OnlyFloats(ak.Array):
        def __awkward_validation__(self) -> None:
            def _content_only_floats(layout, **kwargs) -> None:
                del kwargs
                if layout.is_numpy:
                    if not np.issubdtype(layout.dtype, np.floating):
                        raise OnlyFloatsError(
                            "OnlyFloats arrays can only contain floating-point numbers."
                        )

            _ = ak.transform(_content_only_floats, self, return_value="none")

    behavior = {
        "OnlyFloats": OnlyFloats,
        ("*", "OnlyFloats"): OnlyFloats,
    }

    # this should pass
    _ = ak.with_parameter([1.0], "__list__", "OnlyFloats", behavior=behavior)

    # this should fail
    with pytest.raises(OnlyFloatsError):
        _ = ak.with_parameter([1], "__list__", "OnlyFloats", behavior=behavior)


def test_field_validation():
    class OnlyAllowedFieldsError(Exception): ...

    class OnlyAllowedFields(ak.Array):
        _allowed_fields = frozenset({"pt", "eta", "phi", "mass"})

        def __awkward_validation__(self) -> None:
            fields = set(self.fields)
            if self._allowed_fields != fields:
                raise OnlyAllowedFieldsError(
                    f"OnlyAllowedFields arrays can only have fields: {self._allowed_fields}, "
                    f"but found fields: {fields}"
                )

    behavior = {
        "OnlyAllowedFields": OnlyAllowedFields,
        ("*", "OnlyAllowedFields"): OnlyAllowedFields,
    }

    # this should pass
    _ = ak.with_parameter(
        [{"pt": 1.0, "eta": 1.0, "phi": 1.0, "mass": 1.0}],
        "__list__",
        "OnlyAllowedFields",
        behavior=behavior,
    )

    # this should fail
    with pytest.raises(
        OnlyAllowedFieldsError, match="OnlyAllowedFields arrays can only have fields"
    ):
        _ = ak.with_parameter(
            [{"pt": 1.0, "eta": 1.0, "phi": 1.0, "mass": 1.0, "energy": 1.0}],
            "__list__",
            "OnlyAllowedFields",
            behavior=behavior,
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak


def test_original_issue():
    form = ak.forms.UnionForm(
        "i8",
        "i64",
        [
            ak.forms.NumpyForm("int64", form_key="numpy1"),
            ak.forms.NumpyForm("int64", form_key="numpy2"),
        ],
        form_key="union",
    )
    buffers = {
        "union-tags": np.array([1, 1], dtype=np.int8),
        "union-index": np.array([0, 1], dtype=np.int64),
        "numpy1-data": np.array([1, 2], dtype=np.int64),
        "numpy2-data": np.array([3, 4], dtype=np.int64),
    }
    array = ak.from_buffers(form, 2, buffers)
    assert ak.almost_equal(array, array)
    assert ak.array_equal(array, array)

    form = ak.forms.UnionForm(
        "i8",
        "i64",
        [
            ak.forms.IndexedOptionForm("i64", ak.forms.EmptyForm(), form_key="nones"),
            ak.forms.UnmaskedForm(ak.forms.NumpyForm("int64", form_key="numpy")),
        ],
        form_key="union",
    )
    virtual_buffers = {
        "union-tags": lambda: np.array([1, 1], dtype=np.int8),
        "union-index": lambda: np.array([0, 1], dtype=np.int64),
        "numpy-data": lambda: np.array([1, 2], dtype=np.int64),
        "nones-index": lambda: np.array([-1], dtype=np.int64),
    }
    eager_buffers = {k: v() for k, v in virtual_buffers.items()}
    eager_array = ak.from_buffers(form, 2, eager_buffers)
    virtual_array = ak.from_buffers(form, 2, virtual_buffers)
    assert ak.almost_equal(eager_array, virtual_array)
    assert ak.array_equal(eager_array, virtual_array)

    from awkward._nplikes.numpy import Numpy
    from awkward._nplikes.virtual import VirtualNDArray

    nplike = Numpy.instance()
    form = ak.forms.UnionForm(
        "i8",
        "i64",
        [
            ak.forms.IndexedOptionForm("i64", ak.forms.EmptyForm(), form_key="nones"),
            ak.forms.UnmaskedForm(ak.forms.NumpyForm("int64", form_key="numpy")),
        ],
        form_key="union",
    )
    virtual_buffers = {
        "union-tags": VirtualNDArray(
            nplike,
            dtype=np.int8,
            shape=(2,),
            generator=lambda: np.array([1, 1], dtype=np.int8),
        ),
        "union-index": VirtualNDArray(
            nplike,
            dtype=np.int64,
            shape=(2,),
            generator=lambda: np.array([0, 1], dtype=np.int64),
        ),
        "numpy-data": VirtualNDArray(
            nplike,
            dtype=np.int64,
            shape=(2,),
            generator=lambda: np.array([1, 2], dtype=np.int64),
        ),
        "nones-index": VirtualNDArray(
            nplike,
            dtype=np.int64,
            shape=(1,),
            generator=lambda: np.array([-1], dtype=np.int64),
        ),
    }
    eager_buffers = {k: v._generator() for k, v in virtual_buffers.items()}
    eager_array = ak.from_buffers(form, 2, eager_buffers)
    virtual_array = ak.from_buffers(form, 2, virtual_buffers)
    assert ak.almost_equal(eager_array, virtual_array)
    assert ak.array_equal(eager_array, virtual_array)


def test_empty_union():
    left = ak.contents.UnionArray(
        ak.index.Index8([]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([]),
        ak.index.Index64([0, 0, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([2, 3, 0, 1, 2]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)


def test_size_one_tags():
    left = ak.contents.UnionArray(
        ak.index.Index8([0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1]),
        ak.index.Index64([0, 0, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)


def test_all_tags_used():
    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 2, 3, 0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([2, 3, 0, 1, 2]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 2, 3, 0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([3, 2, 0, 1, 2]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 2, 3, 0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([2, 3, 0, 1, 2]),
        ak.index.Index64([0, 0, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)


def test_unused_tags():
    left = ak.contents.UnionArray(
        ak.index.Index8([1, 2, 2, 1, 1]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([3, 0, 0, 3, 3]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([1, 2, 3, 1, 2]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([3, 0, 1, 3, 0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([1, 2, 3, 1, 2]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([3, 1, 0, 3, 1]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 0, 0, 0, 0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([2, 2, 2, 2, 2]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([1, 1, 1, 1, 1]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([3, 3, 3, 3, 3]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 0, 0, 0, 0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([3, 3, 3, 3, 3]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([1, 1, 1, 1, 1]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([0, 0, 0, 0, 0]),
        ak.index.Index64([0, 1, 0, 0, 1]),
        [
            ak.contents.NumpyArray(np.array([0, 1, 0, 1], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([0, 1], dtype=np.dtype("M8[s]"))),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right)
    assert not ak.array_equal(left, right)
    assert not ak.almost_equal(right, left)
    assert not ak.array_equal(right, left)


def test_castable_dtypes():
    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right, dtype_exact=True)
    assert ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert ak.almost_equal(right, left, dtype_exact=True)
    assert ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert not ak.almost_equal(left, right, dtype_exact=False)
    assert not ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert not ak.almost_equal(right, left, dtype_exact=False)
    assert not ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float32)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float32)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right, dtype_exact=True)
    assert ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert ak.almost_equal(right, left, dtype_exact=True)
    assert ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float32)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float32)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right, dtype_exact=True)
    assert ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert ak.almost_equal(right, left, dtype_exact=True)
    assert ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)


def test_one_has_more_types():
    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([True, False], dtype=np.bool_)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 2, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([1, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([True, False], dtype=np.bool_)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right)
    assert ak.array_equal(left, right)
    assert ak.almost_equal(right, left)
    assert ak.array_equal(right, left)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 2, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([1, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([True, False], dtype=np.bool_)),
            ak.contents.NumpyArray(np.array([True, False], dtype=np.bool_)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 2, 3]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([1, 1, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([True, False], dtype=np.bool_)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 2, 3]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
        ],
    )

    right = ak.contents.UnionArray(
        ak.index.Index8([1, 1, 1, 2]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([True, False], dtype=np.bool_)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert not ak.almost_equal(left, right, dtype_exact=False)
    assert not ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert not ak.almost_equal(right, left, dtype_exact=False)
    assert not ak.array_equal(right, left, dtype_exact=False)


def test_different_integer_sizes():
    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int8)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int16)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int16)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int8)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right, dtype_exact=True)
    assert ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert ak.almost_equal(right, left, dtype_exact=True)
    assert ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int8)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int16)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int8)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int32)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int16)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)


def test_unsigned_integers():
    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint8)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint16)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint16)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint8)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right, dtype_exact=True)
    assert ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert ak.almost_equal(right, left, dtype_exact=True)
    assert ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint32)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint64)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint64)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint32)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right, dtype_exact=True)
    assert ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert ak.almost_equal(right, left, dtype_exact=True)
    assert ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint8)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint32)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint16)),
            ak.contents.NumpyArray(np.array([1, 2], dtype=np.uint64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)


def test_complex_numbers():
    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1 + 2j, 3 + 4j], dtype=np.complex64)),
            ak.contents.NumpyArray(np.array([5 + 6j, 7 + 8j], dtype=np.complex128)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([5 + 6j, 7 + 8j], dtype=np.complex128)),
            ak.contents.NumpyArray(np.array([1 + 2j, 3 + 4j], dtype=np.complex64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert ak.almost_equal(left, right, dtype_exact=True)
    assert ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert ak.almost_equal(right, left, dtype_exact=True)
    assert ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1 + 2j, 3 + 4j], dtype=np.complex64)),
            ak.contents.NumpyArray(np.array([5 + 6j, 7 + 8j], dtype=np.complex128)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([5 + 6j, 7 + 8j], dtype=np.complex128)),
            ak.contents.NumpyArray(np.array([1 + 2j, 3 + 4j], dtype=np.complex64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert not ak.almost_equal(left, right, dtype_exact=False)
    assert not ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert not ak.almost_equal(right, left, dtype_exact=False)
    assert not ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float32)),
            ak.contents.NumpyArray(np.array([3.0, 4.0], dtype=np.float64)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([3 + 0j, 4 + 0j], dtype=np.complex128)),
            ak.contents.NumpyArray(np.array([1 + 0j, 2 + 0j], dtype=np.complex64)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert not ak.almost_equal(left, right, dtype_exact=False)
    assert not ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert not ak.almost_equal(right, left, dtype_exact=False)
    assert not ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([3 + 0j, 4 + 0j], dtype=np.complex64)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([1, 0, 1, 0]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([3 + 0j, 4 + 0j], dtype=np.complex128)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float32)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert ak.almost_equal(left, right, dtype_exact=False)
    assert ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert ak.almost_equal(right, left, dtype_exact=False)
    assert ak.array_equal(right, left, dtype_exact=False)

    left = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float64)),
            ak.contents.NumpyArray(np.array([3 + 0j, 4 + 0j], dtype=np.complex64)),
        ],
    )
    right = ak.contents.UnionArray(
        ak.index.Index8([0, 1, 0, 1]),
        ak.index.Index64([0, 1, 1, 0]),
        [
            ak.contents.NumpyArray(np.array([3 + 0j, 4 + 0j], dtype=np.complex128)),
            ak.contents.NumpyArray(np.array([1.0, 2.0], dtype=np.float32)),
        ],
    )
    assert ak.almost_equal(left, left)
    assert ak.array_equal(left, left)
    assert ak.almost_equal(right, right)
    assert ak.array_equal(right, right)
    assert not ak.almost_equal(left, right, dtype_exact=True)
    assert not ak.array_equal(left, right, dtype_exact=True)
    assert not ak.almost_equal(left, right, dtype_exact=False)
    assert not ak.array_equal(left, right, dtype_exact=False)
    assert not ak.almost_equal(right, left, dtype_exact=True)
    assert not ak.array_equal(right, left, dtype_exact=True)
    assert not ak.almost_equal(right, left, dtype_exact=False)
    assert not ak.array_equal(right, left, dtype_exact=False)
